# DMG Sailthru Content API WordPress Plugin

Sends requests to the Sailthru Content API when posts are created/updated/deleted in WordPress.

Uses the bundled Sailthru PHP SDK (in the `/includes/` dir) if the `Sailthru_Client` class cannot be found.

## Installation

1. Install and activate this plugin.
2. Provide your Sailthu **API Secret** and **API Key**, using either:
   - The [`dmg_sailthru_content_api_auth_key`](#dmg_sailthru_content_api_auth_key) and [`dmg_sailthru_content_api_auth_secret`](#dmg_sailthru_content_api_auth_secret) filters documented below.
   - `Settings > DMG Sailthru Content API` _(if this page is hidden from the admin navbar then the values have already been provided in code using filters)_.

The plugin will now sync the posts you publish/update in WordPress to Sailthru.

## Customisation

To customise the data/post types sent to Sailthru use the [Filters](#Filters) documented below.

For general information on actions and filters refer to the [Hooks](https://developer.wordpress.org/plugins/hooks/) documentation on wordpress.org.

## Testing Sailthru API calls

The plugin makes REST API calls to the Sailthru API to modify `content` entities.

Results of the calls can be viewed in JSON `GET` requests on the Sailthru site:

- Visit https://my.sailthru.com/api/test
- Set the fields:
  - `Method`: `GET`
  - `Action`: `content`
  - `Params`: Optionally target a Post using either:
    - `{"key":"sku","id":"your_id_here"}` (use the Post ID)
    - `{"url":"your_url_here"}` (use the permalink)

If your content is not appearing in Sailthru use the [`dmg_sailthru_content_api_exception`](#dmg_sailthru_content_api_exception) action to log the error messages.

## PHPUnit Tests

[Initialize the testing environment locally](https://make.wordpress.org/cli/handbook/misc/plugin-unit-tests/#3-initialize-the-testing-environment-locally).

Install dev-dependencies:

```bash
composer install
```

Run the tests:

```bash
./vendor/bin/phpunit
```

If the tests don't run you may have tmp directory issues on your machine.

To resolve the tmp directory issues clone [wordpress-develop](https://github.com/WordPress/wordpress-develop) into a known directory and set the `WP_TESTS_DIR` path in your `.bash_profile`, for example:

```bash
export WP_TESTS_DIR="$HOME/wordpress-develop/tests/phpunit"
```

## Actions

### dmg_sailthru_content_api_exception

```php
/**
 * Sailthru Content API Exception hook.
 *
 * Fired when a Sailthru Content API request throws an Exception.
 *
 * If content is not being synced to Sailthru this action can help you to
 * determine why.
 *
 * @param \Exception $exception Exception
 *
 * @return void
 */
function metro_sailthru_content_api_exception( \Exception $exception ): void {
	// Send errors to a Slack channel.
	$payload = [
		'channel'    => '#sailthru-alerts',
		'username'   => 'wp-bot',
		'text'       => (string) $exception,
		'icon_emoji' => ':siren:',
	];

	wp_safe_remote_post(
		get_option( 'slack_api_url' ),
		[
			'body'    => wp_json_encode( $payload ),
			'blocker' => false,
		]
	);
}
add_action( 'dmg_sailthru_content_api_exception', 'metro_sailthru_content_api_exception' );
```

## Filters

### dmg_sailthru_content_api_post_types

```php
/**
 * Set the post types to send to the Sailthru Content API.
 *
 * @param array $post_types Post types.
 *
 * @return array Post types.
 */
function metro_sailthru_content_api_post_types( array $post_types ): array {
	// Also send Pages to the Sailthru Content API.
	$post_types[] = 'page';

	return $post_types;
}
add_filter( 'dmg_sailthru_content_api_post_types', 'metro_sailthru_content_api_post_types' );
```

### dmg_sailthru_content_api_auth_key

```php
/**
 * Set the Sailthru Content API key.
 *
 * @param string $api_key Sailthru API key.
 *
 * @return string Sailthru API key.
 */
function metro_sailthru_content_api_auth_key( string $api_key ): string {
	// Use a WordPress setting for Sailthru API key.
	return get_option( 'sailthru_api_key' );
}
add_filter( 'dmg_sailthru_content_api_auth_key', 'metro_sailthru_content_api_auth_key' );
```

### dmg_sailthru_content_api_auth_secret

```php
/**
 * Set the Sailthru Content API secret.
 *
 * @param string $api_secret Sailthru API secret.
 *
 * @return string Sailthru API secret.
 */
function metro_sailthru_content_api_auth_secret( string $api_secret ): string {
	// Use a WordPress setting for Sailthru API secret.
	return get_option( 'sailthru_api_secret' );
}
add_filter( 'dmg_sailthru_content_api_auth_secret', 'metro_sailthru_content_api_auth_secret' );
```

### dmg_sailthru_content_api_data

```php
/**
 * Filter the entire Sailthru Content API payload.
 *
 * Note that this filter is applied after `dmg_sailthru_content_api_data_vars`.
 *
 * @param array    $data Current `data` array.
 * @param \WP_Post $post Post object.
 *
 * @return array Updated `data` array.
 */
function metro_sailthru_content_api_data( array $data, \WP_Post $post ): array {
	// Add a new property to the data array.
	$data['my_new_property'] = 'my-new-value';

	return $data;
}
add_filter( 'dmg_sailthru_content_api_data', 'metro_sailthru_content_api_data', 10, 2 );
```

### dmg_sailthru_content_api_data_vars

```php
/**
 * Filter the `vars` key of the Sailthru Content API payload.
 *
 * Note that this filter is applied before `dmg_sailthru_content_api_data`.
 *
 * @param array    $vars Current `data.vars` array.
 * @param \WP_Post $post Post object.
 *
 * @return array Updated `data.vars` array.
 */
function metro_sailthru_content_api_data_vars( array $vars, \WP_Post $post ): array {
	// Set the `image_16_9` and `image_4_3` vars.
	$image_id = get_post_thumbnail_id( $post->ID );
	$vars['image_16_9'] = (string) wp_get_attachment_image_url( $image_id, 'landscape_16_9_large' );
	$vars['image_4_3']  = (string) wp_get_attachment_image_url( $image_id, 'landscape_4_3_large' );

	return $vars;
}
add_filter( 'dmg_sailthru_content_api_data_vars', 'metro_sailthru_content_api_data_vars', 10, 2 );
```
