<?php

namespace DMG\Sailthru_Content_API;

/**
 * Class Settings.
 *
 * Handles the options and settings screen.
 *
 * @package DMG\Sailthru_Content_API
 */
class Settings {

	/**
	 * Option group.
	 *
	 * @var string
	 */
	const OPTION_GROUP = 'dmg_sailthru_content_api';

	/**
	 * Option name for the Sailthru API Key.
	 *
	 * @var string
	 */
	const OPTION_API_KEY = 'dmg_sailthru_content_api_auth_key';

	/**
	 * Option name for the Sailthru API Secret.
	 *
	 * @var string
	 */
	const OPTION_API_SECRET = 'dmg_sailthru_content_api_auth_secret';

	/**
	 * Settings page.
	 *
	 * @var string
	 */
	const SETTINGS_PAGE = 'dmg-sailthru-content-api-settings';

	/**
	 * Settings section ID for the "Authentication" section.
	 *
	 * @var string
	 */
	const SETTINGS_SECTION_AUTHENTICATION = 'authentication';

	/**
	 * Init.
	 *
	 * @return void
	 */
	public static function init(): void {
		add_action( 'admin_init', [ __CLASS__, 'add_settings_sections' ] );
		add_action( 'admin_init', [ __CLASS__, 'register_settings' ] );
		add_action( 'admin_init', [ __CLASS__, 'add_settings_fields' ] );
		add_action( 'admin_menu', [ __CLASS__, 'add_options_page' ] );
	}

	/**
	 * Get the Sailthru API Key from the filter or the options,
	 * prioritising the filter value even if it returns empty.
	 *
	 * @throws \Exception On missing/empty API Key.
	 *
	 * @return string Sailthru API Key.
	 */
	public static function get_api_key(): string {
		if ( has_filter( 'dmg_sailthru_content_api_auth_key' ) ) {
			$api_key = apply_filters( 'dmg_sailthru_content_api_auth_key', '' );
		} else {
			$api_key = get_option( Settings::OPTION_API_KEY, '' );
		}

		if ( empty( $api_key ) ) {
			throw new \Exception( 'Sailthru API Key is empty.' );
		}

		return $api_key;
	}

	/**
	 * Get the Sailthru API Secret from the filter or the options,
	 * prioritising the filter value even if it returns empty.
	 *
	 * @throws \Exception On missing/empty API Secret.
	 *
	 * @return string Sailthru API Secret.
	 */
	public static function get_api_secret(): string {
		if ( has_filter( 'dmg_sailthru_content_api_auth_secret' ) ) {
			$api_secret = apply_filters( 'dmg_sailthru_content_api_auth_secret', '' );
		} else {
			$api_secret = get_option( Settings::OPTION_API_SECRET, '' );
		}

		if ( empty( $api_secret ) ) {
			throw new \Exception( 'Sailthru API Secret is empty.' );
		}

		return $api_secret;
	}

	/**
	 * Add settings sections.
	 *
	 * @return void
	 */
	public static function add_settings_sections(): void {
		add_settings_section(
			self::SETTINGS_SECTION_AUTHENTICATION,
			__( 'Authentication', 'dmg-sailthru-content-api' ),
			'__return_false',
			self::SETTINGS_PAGE
		);
	}

	/**
	 * Register the setting fields.
	 *
	 * @return void
	 */
	public static function register_settings(): void {
		// "API Key" setting.
		register_setting(
			self::OPTION_GROUP,
			self::OPTION_API_KEY,
			[
				'type'              => 'string',
				'sanitize_callback' => 'sanitize_text_field',
				'show_in_rest'      => false,
				'default'           => '',
			]
		);

		// "API Secret" setting.
		register_setting(
			self::OPTION_GROUP,
			self::OPTION_API_SECRET,
			[
				'type'              => 'string',
				'sanitize_callback' => 'sanitize_text_field',
				'show_in_rest'      => false,
				'default'           => '',
			]
		);
	}

	/**
	 * Register the setting fields
	 *
	 * @return void
	 */
	public static function add_settings_fields(): void {
		// "API Key" field.
		add_settings_field(
			self::OPTION_API_KEY,
			__( 'Sailthru API Key', 'dmg-sailthru-content-api' ),
			[
				__CLASS__,
				'field_html_api_key',
			],
			self::SETTINGS_PAGE,
			self::SETTINGS_SECTION_AUTHENTICATION
		);

		// "API Secret" field.
		add_settings_field(
			self::OPTION_API_SECRET,
			__( 'Sailthru API Secret', 'dmg-sailthru-content-api' ),
			[
				__CLASS__,
				'field_html_api_secret',
			],
			self::SETTINGS_PAGE,
			self::SETTINGS_SECTION_AUTHENTICATION
		);
	}

	/**
	 * Callback for the "API Key" field HTML.
	 *
	 * @return void
	 */
	public static function field_html_api_key(): void {
		$has_filter = has_filter( 'dmg_sailthru_content_api_auth_key' );

		$value = '';
		if ( ! $has_filter ) {
			$value = get_option( self::OPTION_API_KEY, '' );
		}
		?>
		<input
			type="password"
			id="<?php echo esc_attr( self::OPTION_API_KEY ); ?>"
			name="<?php echo esc_attr( self::OPTION_API_KEY ); ?>"
			value="<?php echo esc_attr( $value ); ?>"
			<?php disabled( $has_filter ); ?>
		/>
		<?php
		if ( $has_filter ) {
			?>
			<p class="description">This value is not editable here.</p>
			<p class="description">It has been set in code using the <code>dmg_sailthru_content_api_auth_key</code> filter.</p>
			<?php
		}
	}

	/**
	 * Callback for the "API Secret" field HTML.
	 *
	 * @return void
	 */
	public static function field_html_api_secret(): void {
		$has_filter = has_filter( 'dmg_sailthru_content_api_auth_secret' );

		$value = '';
		if ( ! $has_filter ) {
			$value = get_option( self::OPTION_API_SECRET, '' );
		}
		?>
		<input
			type="password"
			id="<?php echo esc_attr( self::OPTION_API_SECRET ); ?>"
			name="<?php echo esc_attr( self::OPTION_API_SECRET ); ?>"
			value="<?php echo esc_attr( $value ); ?>"
			<?php disabled( $has_filter ); ?>
		/>
		<?php
		if ( $has_filter ) {
			?>
			<p class="description">This value is not editable here.</p>
			<p class="description">It has been set in code using the <code>dmg_sailthru_content_api_auth_secret</code> filter.</p>
			<?php
		}
	}

	/**
	 * Add menu item to the WordPress admin menu, at
	 * Settings > DMG Sailthru Content API.
	 *
	 * @return void
	 */
	public static function add_options_page(): void {
		// If a filter has been added then we prioritise it's value even if it's empty.
		$has_api_key_filter    = has_filter( 'dmg_sailthru_content_api_auth_key' );
		$has_api_secret_filter = has_filter( 'dmg_sailthru_content_api_auth_secret' );

		// Bail if all API creds are provided by filters.
		if ( $has_api_key_filter && $has_api_secret_filter ) {
			return;
		}

		add_options_page(
			__( 'Sailthru Content API', 'speechkit' ),
			__( 'Sailthru Content API', 'speechkit' ),
			'manage_options',
			self::SETTINGS_PAGE,
			[ __CLASS__, 'form_html' ]
		);
	}

	/**
	 * The form HTML for the settings page.
	 *
	 * @return void
	 */
	public static function form_html(): void {
		?>
		<div class="wrap">
			<form
				id="<?php echo esc_attr( self::SETTINGS_PAGE ); ?>"
				action="<?php echo esc_url( admin_url( 'options.php' ) ); ?>"
				method="post"
			>
				<h1>
					<?php esc_html_e( 'Sailthru Content API', 'speechkit' ); ?>
				</h1>
				<?php
				settings_fields( self::OPTION_GROUP );
				do_settings_sections( self::SETTINGS_PAGE );
				submit_button();
				?>
			</form>
		</div>
		<?php
	}
}

Settings::init();
