<?php

namespace DMG\Sailthru_Content_API;

/**
 * Class Sync.
 *
 * Syncs posts between WordPress and the Sailthru API when the
 * posts are published/updated.
 *
 * @package DMG\Sailthru_Content_API
 */
class Sync {

	/**
	 * Cron event name.
	 *
	 * @var string
	 */
	const CRON_NAME = 'dmg_sailthru_content_api_sync';

	/**
	 * Cron event delay in seconds.
	 *
	 * @var int
	 */
	const CRON_DELAY = 10;

	/**
	 * Default post types to sync.
	 *
	 * @var array
	 */
	const DEFAULT_POST_TYPES = [ 'post' ];

	/**
	 * Post meta name for storing Sailthru API responses.
	 *
	 * @var string
	 */
	const META_NAME = 'dmg_sailthru_content_api_response';

	/**
	 * Add actions to perform the content sync through cron tasks.
	 *
	 * @return void
	 */
	public static function init(): void {
		add_action( 'transition_post_status', [ __CLASS__, 'maybe_schedule_event_on_status_change' ], 10, 3 );
		add_action( 'wp_after_insert_post', [ __CLASS__, 'maybe_schedule_event_on_update' ], 10, 4 );
		add_action( self::CRON_NAME, [ __CLASS__, 'maybe_content_api_sync' ], 10, 2 );
	}

	/**
	 * Maybe schedule a cron event when a post status transitions.
	 *
	 * @param string   $new_status The new post status.
	 * @param string   $old_status The old post status.
	 * @param \WP_Post $post       The post object.
	 *
	 * @return string|false The action of the scheduled event, or false if it wasn't scheduled.
	 */
	public static function maybe_schedule_event_on_status_change( string $new_status, string $old_status, \WP_Post $post ): string|false {
		// Check this isn't a revision or autosave.
		if ( wp_is_post_autosave( $post ) || wp_is_post_revision( $post ) ) {
			return false;
		}

		// Check we've got valid post objects.
		if ( ! is_a( $post, 'WP_Post' ) ) {
			return false;
		}

		// Check we sync this post type.
		$post_types = apply_filters( 'dmg_sailthru_content_api_post_types', self::DEFAULT_POST_TYPES );
		if ( ! in_array( $post->post_type, $post_types, true ) ) {
			return false;
		}

		$action = false;

		if ( 'publish' === $new_status && 'publish' !== $old_status ) {
			// Post is being published.
			$action = 'create';
		} elseif ( 'publish' !== $new_status && 'publish' === $old_status ) {
			// Post is being unpublished.
			$action = 'delete';
		}

		if ( $action ) {
			// Schedule the event (note that scheduling may still fail e.g. if API creds are missing)
			self::schedule_event( $post->ID, $action );
		}

		return $action;
	}

	/**
	 * Maybe schedule a cron event when a post is updated.
	 *
	 * @param int           $post_id     Post ID.
	 * @param \WP_Post      $post        Post object.
	 * @param bool          $update      Whether this is an existing post being updated.
	 * @param null|\WP_Post $post_before Null for new posts, the WP_Post object prior
	 *                                  to the update for updated posts.
	 *
	 * @return string|false The action of the scheduled event, or false if it wasn't scheduled.
	 */
	public static function maybe_schedule_event_on_update( int $post_id, \WP_Post $post, bool $update, null|\WP_Post $post_before ): string|false {
		// Check an existing post is being updated.
		if ( true !== $update ) {
			return false;
		}

		// Check this isn't a revision or autosave.
		if ( wp_is_post_autosave( $post ) || wp_is_post_revision( $post ) ) {
			return false;
		}

		// Check we've got valid post objects.
		if ( ! is_a( $post, 'WP_Post' ) || ! is_a( $post_before, 'WP_Post' ) ) {
			return false;
		}

		// Check the update has been made to a published post.
		if ( 'publish' !== $post->post_status || 'publish' !== $post_before->post_status ) {
			return false;
		}

		// Check the update has been made to a published post.
		if ( 'future' === $post->post_status ) {
			return false;
		}

		// Check we sync this post type.
		$post_types = apply_filters( 'dmg_sailthru_content_api_post_types', self::DEFAULT_POST_TYPES );
		if ( ! in_array( $post->post_type, $post_types, true ) ) {
			return false;
		}

		$action = 'update';

		// Schedule the event (note that scheduling may still fail e.g. if API creds are missing)
		self::schedule_event( $post->ID, $action );

		return $action;
	}

	/**
	 * Schedule the content API sync event for a post.
	 *
	 * Errors throw an Exception. Hook into the `dmg_sailthru_content_api_exception`
	 * action to handle them.
	 *
	 * @throws \Exception Exception thrown during the cron job should be handled
	 *                    with the `dmg_sailthru_content_api_exception` action.
	 *
	 * @param int    $post_id Post ID.
	 * @param string $action  Action to perform (create|update|delete).
	 *
	 * @return void
	 */
	public static function schedule_event( int $post_id, string $action ): void {
		try {
			// Check the action is valid.
			if ( ! in_array( $action, [ 'create', 'update', 'delete' ], true ) ) {
				throw new \Exception( 'Unexpected action argument for cron task' );
			}

			// Check for valid API creds (Exceptions are thrown if any are missing)
			\DMG\Sailthru_Content_API\Settings::get_api_key();
			\DMG\Sailthru_Content_API\Settings::get_api_secret();

			// Check if the cron has already been scheduled.
			if ( wp_next_scheduled( self::CRON_NAME, [ $post_id, $action ] ) ) {
				return;
			}

			// Schedule the cron task.
			wp_schedule_single_event( time() + self::CRON_DELAY, self::CRON_NAME, [ $post_id, $action ] );
		} catch ( \Exception $e ) {
			// Action for all sync errors.
			do_action( 'dmg_sailthru_content_api_exception', $e );
		}
	}

	/**
	 * Perform the Sailthru content API sync request.
	 *
	 * Errors throw an Exception. Hook into the `dmg_sailthru_content_api_exception`
	 * action to handle them.
	 *
	 * @throws \Exception Exception thrown during the cron job should be handled
	 *                    with the `dmg_sailthru_content_api_exception` action.
	 *
	 * @param int    $post_id Post ID.
	 * @param string $action  Action to perform (create|update|delete).
	 *
	 * @return void
	 */
	public static function maybe_content_api_sync( int $post_id, string $action ): void {
		try {
			$api_key    = \DMG\Sailthru_Content_API\Settings::get_api_key();
			$api_secret = \DMG\Sailthru_Content_API\Settings::get_api_secret();

			$sailthru_client = new \Sailthru_Client( $api_key, $api_secret );

			if ( in_array( $action, [ 'create', 'update' ], true ) ) {
				$data = self::extended_data_payload( $post_id );

				// The `url` is the primary key in Sailthru no matter what the `sku` is, so only set `url` once.
				if ( ! self::content_exists_in_sailthru( $post_id, $sailthru_client ) ) {
					$data['url'] = get_permalink( $post_id );
				}

				// POST request.
				$response = $sailthru_client->apiPost( 'content', $data );

				if ( self::is_error_response( $response ) ) {
					throw new \Exception( self::get_response_error_message( $response ) );
				}

				update_post_meta( $post_id, self::META_NAME, $response );
			} elseif ( 'delete' === $action ) {
				// Only DELETE if the post has response data from a previous Sailthru POST request.
				$prev_response = get_post_meta( $post_id, self::META_NAME, true );
				if ( empty( $prev_response ) ) {
					return;
				}

				/**
				 * For a DELETE we can't reliably distinguish "Content not found" error codes from
				 * other error codes (which also use code `99`), so we check to see whether
				 * the content exists in Sailthru first with a GET before attempting the DELETE.
				 */
				if ( ! self::content_exists_in_sailthru( $post_id, $sailthru_client ) ) {
					return;
				}

				$data = self::minimal_data_payload( $post_id );

				// DELETE request.
				$response = $sailthru_client->apiDelete( 'content', $data );

				if ( self::is_error_response( $response ) ) {
					throw new \Exception( self::get_response_error_message( $response ) );
				}

				delete_post_meta( $post_id, self::META_NAME );
			}
		} catch ( \Exception $e ) {
			// Action for all sync errors.
			do_action( 'dmg_sailthru_content_api_exception', $e );
		}
	}

	/**
	 * Check whether a Sailthru API response is an error reponse.
	 *
	 * This check is based on the "Example Call" in the PHP5 Library SDK docs.
	 *
	 * @link https://getstarted.sailthru.com/developers/api-client/php5/#Example_Call
	 *
	 * @throws \Exception Sailthru API error code and message.
	 *
	 * @param mixed $response Sailthru API response.
	 *
	 * @return bool
	 */
	public static function is_error_response( mixed $response ): bool {
		return is_array( $response ) && isset( $response['error'] );
	}

	/**
	 * Get the error message from a Sailthru response.
	 *
	 * @param mixed $response Sailthru API response.
	 *
	 * @return string Response error message, or an empty string if no error message was found.
	 */
	public static function get_response_error_message( mixed $response ): string {
		if ( is_array( $response ) && isset( $response['errormsg'] ) ) {
			return $response['errormsg'];
		}

		return '';
	}

	/**
	 * Perform a GET request to the Sailthru content API to check whether
	 * content already exists.
	 *
	 * We do this before making a POST request to the Sailthru content API, because
	 * the `url` is actually the primary key in Sailthru, not the `sku` (Post ID).
	 *
	 * Since the permalink may change in WordPress we only set `url` once, when this
	 * GET request returns `false`.
	 *
	 * @param int              $post_id         Post ID.
	 * @param \Sailthru_Client $sailthru_client Sailthru client.
	 *
	 * @return boolean
	 */
	public static function content_exists_in_sailthru( int $post_id, \Sailthru_Client $sailthru_client ): bool {
		try {
			$data = self::minimal_data_payload( $post_id );

			// Check ID exists in Sailthru (should throw a Sailthru_Client_Exception if it doesn't).
			$response = $sailthru_client->apiGet( 'content', $data );

			// Check response body isn't an error.
			if ( ! self::is_error_response( $response ) ) {
				return true;
			}
		} catch ( \Sailthru_Client_Exception $e ) {
			return false;
		}

		return false;
	}

	/**
	 * Get the minimal post data payload to be sent in Sailthru Content API requests.
	 *
	 * This consts of only the `key` and `id` params, and is used only for
	 * GET and DELETE requests.
	 *
	 * @param int $post_id Post ID.
	 *
	 * @return array Content API payload.
	 */
	public static function minimal_data_payload( int $post_id ): array {
		return [
			'key' => 'sku',
			'id'  => (string) $post_id,
		];
	}

	/**
	 * Get the extended post data payload to be sent in Sailthru Content API requests.
	 *
	 * This includes the post title, description, images and custom vars, and is
	 * used for PUT/POST requests.
	 *
	 * @param int $post_id Post ID.
	 *
	 * @return array|false Content API payload, or false on failure.
	 */
	public static function extended_data_payload( int $post_id ): array|false {
		$post = get_post( $post_id );

		if ( ! is_a( $post, 'WP_Post' ) ) {
			return false;
		}

		$vars = apply_filters( 'dmg_sailthru_content_api_data_vars', [], $post );

		$data = [
			'key'         => 'sku',
			'id'          => (string) $post_id,
			'title'       => $post->post_title,
			'description' => $post->post_excerpt,
			'images'      => self::get_images( $post_id ),
			'author'      => self::get_authors( $post_id ),
			'tags'        => self::get_tag_slugs( $post_id ),
			'keys'        => [
				'sku' => (string) $post_id,
			],
			'vars'        => $vars,
		];

		// Allow sites to override $data with a filter.
		$data = apply_filters( 'dmg_sailthru_content_api_data', $data, $post );

		return $data;
	}

	/**
	 * Get images for the API request body.
	 *
	 * @param int $post_id Post ID.
	 *
	 * @return array Associative array of images.
	 */
	public static function get_images( int $post_id ): array {
		$image_id = get_post_thumbnail_id( $post_id );

		$images = [
			'full'  => [
				'url' => (string) wp_get_attachment_image_url( $image_id, 'full' ),
			],
			'thumb' => [
				'url' => (string) wp_get_attachment_image_url( $image_id, 'post-thumbnail' ),
			],
		];

		return $images;
	}

	/**
	 * Get authors for the API request body.
	 *
	 * @param int $post_id Post ID.
	 *
	 * @return string Comma-separated list of authors.
	 */
	public static function get_authors( int $post_id ): string {
		if ( function_exists( 'get_coauthors' ) ) {
			$coauthors       = \get_coauthors( $post_id );
			$coauthors_names = is_array( $coauthors ) ? wp_list_pluck( $coauthors, 'display_name' ) : [];
			$authors         = implode( ', ', $coauthors_names );
		} else {
			$author_id = get_post_field( 'post_author', $post_id );
			$authors   = get_the_author_meta( 'display_name', $author_id );
		}

		return $authors;
	}

	/**
	 * Get tag slugs for the API request body.
	 *
	 * @param int $post_id Post ID.
	 *
	 * @return string Comma-separated list of tag slugs.
	 */
	public static function get_tag_slugs( int $post_id ): string {
		$tag_slugs = [];

		$tags = get_the_tags( $post_id );

		if ( is_array( $tags ) ) {
			$tag_slugs = wp_list_pluck( $tags, 'slug' );
		}

		return implode( ',', $tag_slugs );
	}
}

Sync::init();
