<?php

use DMG\Sailthru_Content_API\Sync;

/**
 * Class Test_Sailthru_Content_API_Sync.
 *
 * Unit tests for Sailthru Content API sync.
 *
 * @covers \DMG\Sailthru_Content_API\Sync
 * @group  dmg-sailthru-content-api
 * @group  dmg-sailthru-content-api-sync
 */
class Test_Sailthru_Content_API_Sync extends WP_UnitTestCase {

	/**
	 * Default post for tests.
	 *
	 * @var \WP_Post|null
	 */
	private \WP_Post|null $post = null;

	/**
	 * Set up.
	 */
	function set_up() {
		parent::set_up();

		add_filter( 'dmg_sailthru_content_api_auth_key', array( __CLASS__, '__return_test_string' ) );
		add_filter( 'dmg_sailthru_content_api_auth_secret', array( __CLASS__, '__return_test_string' ) );

		$this->post = $this->factory()->post->create_and_get(
			[
				'post_content' => file_get_contents( __DIR__ . '/mocks/paragraphs.txt' ),
				'post_status'  => 'publish',
			]
		);
	}

	/**
	 * Tear down.
	 */
	function tear_down() {
		parent::tear_down();
	}

	/**
	 * Returns a test string for testing the API Key and API secret filters.
	 */
	public static function __return_test_string() {
		return 'test';
	}

	/**
	 * Test cron has expected action for status changes.
	 *
	 * @dataProvider maybe_schedule_event_on_status_change_provider
	 */
	function test_maybe_schedule_event_on_status_change( $expect, $new_status, $old_status ) {
		$this->assertSame( $expect, Sync::maybe_schedule_event_on_status_change( $new_status, $old_status, $this->post ) );
	}

	/**
	 * Expected actions for post status transitions.
	 */
	function maybe_schedule_event_on_status_change_provider() {
		return [
			'draft->publish'   => [ 'create', 'publish', 'draft' ],
			'future->publish'  => [ 'create', 'publish', 'future' ],
			'pending->publish' => [ 'create', 'publish', 'pending' ],
			'private->publish' => [ 'create', 'publish', 'private' ],
			'trash->publish'   => [ 'create', 'publish', 'trash' ],
			'publish->draft'   => [ 'delete', 'draft', 'publish' ],
			'publish->future'  => [ 'delete', 'future', 'publish' ],
			'publish->pending' => [ 'delete', 'pending', 'publish' ],
			'publish->private' => [ 'delete', 'private', 'publish' ],
			'publish->trash'   => [ 'delete', 'trash', 'publish' ],
			'publish->publish' => [ false, 'publish', 'publish' ],
		];
	}

	/**
	 * Test cron has expected action for updates to published posts.
	 *
	 * @dataProvider maybe_schedule_event_on_update_provider
	 */
	function test_maybe_schedule_event_on_update( $expect, $update, $new_status, $old_status ) {
		$post = $this->factory()->post->create_and_get(
			[
				'post_content' => file_get_contents( __DIR__ . '/mocks/paragraphs.txt' ),
				'post_status'  => $new_status,
				'post_date'    => $new_status === 'future' ? '2525-02-03' : null,
			]
		);

		$post_before = $this->factory()->post->create_and_get(
			[
				'post_content' => file_get_contents( __DIR__ . '/mocks/paragraphs.txt' ),
				'post_status'  => $old_status,
				'post_date'    => $old_status === 'future' ? '2525-02-03' : null,
			]
		);

		$this->assertSame( $expect, Sync::maybe_schedule_event_on_update( $post->ID, $post, $update, $post_before ) );
	}

	/**
	 * Expected actions for post status transitions.
	 */
	function maybe_schedule_event_on_update_provider() {
		return [
			'future->future'   => [ false, true, 'future', 'future' ],
			'draft->publish'   => [ false, true, 'publish', 'draft' ],
			'future->publish'  => [ false, true, 'publish', 'future' ],
			'pending->publish' => [ false, true, 'publish', 'pending' ],
			'private->publish' => [ false, true, 'publish', 'private' ],
			'trash->publish'   => [ false, true, 'publish', 'trash' ],
			'publish->draft'   => [ false, true, 'draft', 'publish' ],
			'publish->future'  => [ false, true, 'future', 'publish' ],
			'publish->pending' => [ false, true, 'pending', 'publish' ],
			'publish->private' => [ false, true, 'private', 'publish' ],
			'publish->trash'   => [ false, true, 'trash', 'publish' ],
			'publish->publish' => [ 'update', true, 'publish', 'publish' ],
			'$update = false'  => [ false, false, 'publish', 'publish' ],
		];
	}

	/**
	 * Test schedule_event() schedules an event for valid actions.
	 *
	 * @dataProvider schedule_event_valid_actions_provider
	 */
	function test_schedule_event_valid_actions( $action ) {
		Sync::schedule_event( $this->post->ID, $action );

		$this->assertIsInt( wp_next_scheduled( Sync::CRON_NAME, [ $this->post->ID, $action ] ) );
	}

	/**
	 * Valid actions for post status transitions.
	 */
	function schedule_event_valid_actions_provider() {
		return [
			[ 'create' ],
			[ 'update' ],
			[ 'delete' ],
		];
	}

	/**
	 * Test schedule_event() does not schedules an event for an valid action.
	 */
	function test_schedule_event__invalid_action() {
		Sync::schedule_event( $this->post->ID, 'invalid-action' );

		$this->assertFalse( wp_next_scheduled( Sync::CRON_NAME, [ $this->post->ID, 'invalid-action' ] ) );
	}

	/**
	 * Test cron scheduling requires an valid action.
	 */
	function test_schedule_event_requires_a_valid_action() {
		$action = new MockAction();
		add_action( 'dmg_sailthru_content_api_exception', [ $action, 'action' ] );

		$this->assertSame( 0, $action->get_call_count() );
		Sync::schedule_event( $this->post->ID, 'invalid-action' );
		$this->assertSame( 1, $action->get_call_count() );

		$args = $action->get_args();
		$this->assertInstanceOf( 'Exception', $args[0][0] );
		$this->assertSame( 'Unexpected action argument for cron task', $args[0][0]->getMessage() );
	}

	/**
	 * Test cron scheduling requires an API Key.
	 */
	function test_schedule_event_without_api_key() {
		remove_filter( 'dmg_sailthru_content_api_auth_key', array( __CLASS__, '__return_test_string' ) );

		$action = new MockAction();
		add_action( 'dmg_sailthru_content_api_exception', [ $action, 'action' ] );

		$this->assertSame( 0, $action->get_call_count() );
		Sync::schedule_event( $this->post->ID, 'create' );
		$this->assertSame( 1, $action->get_call_count() );

		$args = $action->get_args();
		$this->assertInstanceOf( 'Exception', $args[0][0] );
		$this->assertSame( 'Sailthru API Key is empty.', $args[0][0]->getMessage() );
	}

	/**
	 * Test cron scheduling requires an API Secret.
	 */
	function test_schedule_event_without_api_secret() {
		remove_filter( 'dmg_sailthru_content_api_auth_secret', array( __CLASS__, '__return_test_string' ) );

		$action = new MockAction();
		add_action( 'dmg_sailthru_content_api_exception', [ $action, 'action' ] );

		$this->assertSame( 0, $action->get_call_count() );
		Sync::schedule_event( $this->post->ID, 'create' );
		$this->assertSame( 1, $action->get_call_count() );

		$args = $action->get_args();
		$this->assertInstanceOf( 'Exception', $args[0][0] );
		$this->assertSame( 'Sailthru API Secret is empty.', $args[0][0]->getMessage() );
	}

	/**
	 * Test cron is scheduled if creds are provided.
	 */
	function test_cron_is_not_scheduled_if_creds_are_provided() {
		$this->assertIsInt( wp_next_scheduled( Sync::CRON_NAME, [ $this->post->ID, 'create' ] ) );
	}

	/**
	 * Test support for other post types can be added.
	 */
	function test_post_types_filter() {
		$args = [
			'post_type'    => 'page',
			'post_content' => file_get_contents( __DIR__ . '/mocks/paragraphs.txt' ),
			'post_status'  => 'publish',
		];

		// Create a new page before adding 'page' to the allowed list.
		$post_id = $this->factory()->post->create( $args );

		$this->assertFalse( wp_next_scheduled( Sync::CRON_NAME, [ $post_id, 'create' ] ) );

		// Add 'page' support.
		add_filter(
			'dmg_sailthru_content_api_post_types',
			function ( $post_types ) {
				$post_types['page'] = 'page';
				return $post_types;
			}
		);

		// Create another page with the same $args.
		$post_id = $this->factory()->post->create( $args );

		$this->assertIsInt( wp_next_scheduled( Sync::CRON_NAME, [ $post_id, 'create' ] ) );
	}

	/**
	 * Test the post transitions that should schedule a Sailthru "create" sync.
	 *
	 * @dataProvider transitions_that_schedule_create_provider
	 */
	public function test_create_sync_scheduling( string $old_status, string $new_status ) {
		$post_id = $this->factory()->post->create(
			[
				'post_content' => file_get_contents( __DIR__ . '/mocks/paragraphs.txt' ),
				'post_status'  => $old_status,
			]
		);

		// Original status should not be scheduled for sync.
		$this->assertFalse( wp_next_scheduled( Sync::CRON_NAME, [ $post_id, 'create' ] ) );

		wp_update_post(
			[
				'ID'          => $post_id,
				'post_status' => $new_status,
			]
		);

		// Status transition should have scheduled a sync.
		$this->assertIsInt( wp_next_scheduled( Sync::CRON_NAME, [ $post_id, 'create' ] ) );
	}

	/**
	 * Transitions that schedule a "create" sync.
	 *
	 * @todo do we need to handle 'future' status to/from transitions?
	 */
	public function transitions_that_schedule_create_provider() {
		return [
			[
				'old_status' => 'draft',
				'new_status' => 'publish',
			],
			[
				'old_status' => 'pending',
				'new_status' => 'publish',
			],
			[
				'old_status' => 'private',
				'new_status' => 'publish',
			],
		];
	}

	/**
	 * Test the post transitions that should schedule a Sailthru "delete" sync.
	 *
	 * @dataProvider transitions_that_schedule_delete_provider
	 */
	public function test_delete_sync_scheduling( string $old_status, string $new_status ) {
		$post_id = $this->factory()->post->create(
			[
				'post_content' => file_get_contents( __DIR__ . '/mocks/paragraphs.txt' ),
				'post_status'  => $old_status,
			]
		);

		// Original status should not be scheduled for sync.
		$this->assertFalse( wp_next_scheduled( Sync::CRON_NAME, [ $post_id, 'delete' ] ) );

		wp_update_post(
			[
				'ID'          => $post_id,
				'post_status' => $new_status,
			]
		);

		// Status transition should have scheduled a sync.
		$this->assertIsInt( wp_next_scheduled( Sync::CRON_NAME, [ $post_id, 'delete' ] ) );
	}

	/**
	 * Transitions that schedule a "delete" sync.
	 */
	public function transitions_that_schedule_delete_provider() {
		return [
			[
				'old_status' => 'publish',
				'new_status' => 'draft',
			],
			[
				'old_status' => 'publish',
				'new_status' => 'pending',
			],
			[
				'old_status' => 'publish',
				'new_status' => 'private',
			],
		];
	}

	/**
	 * Test the post transitions that don't schedule any Sailthru "create"/"delete" syncs.
	 *
	 * @dataProvider transitions_that_dont_schedule_provider
	 */
	public function test_transitions_that_dont_schedule_a_sync( string $old_status, string $new_status ) {
		$post_id = $this->factory()->post->create(
			[
				'post_title'   => 'Original title',
				'post_content' => file_get_contents( __DIR__ . '/mocks/paragraphs.txt' ),
				'post_status'  => $old_status,
			]
		);

		// Original status should not be scheduled for sync.
		$this->assertFalse( wp_next_scheduled( Sync::CRON_NAME, [ $post_id, 'delete' ] ) );

		wp_update_post(
			[
				'ID'          => $post_id,
				'post_title'  => 'Updated title',
				'post_status' => $new_status,
			]
		);

		// Status transition should also not have scheduled a sync.
		$this->assertFalse( wp_next_scheduled( Sync::CRON_NAME, [ $post_id, 'delete' ] ) );
	}

	/**
	 * Transitions that don't schedule any syncs.
	 */
	public function transitions_that_dont_schedule_provider() {
		return [
			[
				'old_status' => 'draft',
				'new_status' => 'draft',
			],
			[
				'old_status' => 'draft',
				'new_status' => 'pending',
			],
			[
				'old_status' => 'draft',
				'new_status' => 'private',
			],
			[
				'old_status' => 'pending',
				'new_status' => 'private',
			],
		];
	}

	/**
	 * Test is_error_response().
	 *
	 * @dataProvider error_response_provider
	 */
	public function test_is_error_response( bool $expect, mixed $response ) {
		$this->assertSame( $expect, Sync::is_error_response( $response ) );
	}

	/**
	 * Error response provider.
	 */
	public function error_response_provider() {
		return [
			[ false, null ],
			[ false, '' ],
			[ false, [] ],
			[ false, [ 'errormsg' => 'Only a message' ] ],
			[
				true,
				[
					'error'    => 99,
					'errormsg' => 'Error message'
				]
			],
		];
	}

	/**
	 * Test get_response_error_message().
	 *
	 * @dataProvider error_response_provider
	 */
	public function test_get_response_error_message( bool $expect, mixed $response ) {
		$this->assertSame( $expect, Sync::is_error_response( $response ) );
	}

	/**
	 * Error message provider.
	 */
	public function response_error_message_provider() {
		return [
			[ '', null ],
			[
				'Error message',
				[
					'error'    => 99,
					'errormsg' => 'Error message'
				]
			],
		];
	}

	/**
	 * Test the minimal data payload for a post.
	 */
	public function test_minimal_data_payload() {
		$post_id = $this->factory()->post->create();

		$expect = [
			'key' => 'sku',
			'id'  => (string) $post_id,
		];

		$payload = Sync::minimal_data_payload( $post_id );

		$this->assertSame( $expect, $payload );
	}

	/**
	 * Test the default Sailthru Content API request data for a post.
	 */
	public function test_default_extended_data_payload() {
		$user_id = $this->factory()->user->create(
			[
				'role'         => 'editor',
				'user_login'   => 'sailthru-editor',
				'display_name' => 'Sailthru Editor',
			]
		);

		$post_id = $this->factory()->post->create(
			[
				'post_title'   => 'Payloads test',
				'post_excerpt' => 'The excerpt.',
				'post_content' => file_get_contents( __DIR__ . '/mocks/paragraphs.txt' ),
				'post_status'  => 'publish',
				'post_author'  => $user_id,
			]
		);

		wp_set_object_terms( $post_id, [ 'tag-1', 'tag-2', 'tag-3' ], 'post_tag', false );

		// Set the featured image
		$attachment_id = $this->factory()->attachment->create_upload_object(
			DIR_TESTDATA . '/images/test-image.jpg',
			$post_id
		);

		$full_size_url = wp_get_attachment_image_url( $attachment_id, 'full' );
		$this->assertNotEmpty( $full_size_url );

		$thumbnail_size_url = wp_get_attachment_image_url( $attachment_id, 'post-thumbnail' );
		$this->assertNotEmpty( $thumbnail_size_url );

		set_post_thumbnail( $post_id, $attachment_id );

		// The JSON payload to send to Sailthru
		$payload = Sync::extended_data_payload( $post_id );

		$expect = [
			'key'         => 'sku',
			'id'          => (string) $post_id,
			'title'       => 'Payloads test',
			'description' => 'The excerpt.',
			'images'      => [
				'full'  => [
					'url' => $full_size_url,
				],
				'thumb' => [
					'url' => $thumbnail_size_url,
				],
			],
			'author'      => 'Sailthru Editor',
			'tags'        => 'tag-1,tag-2,tag-3',
			'keys'        => [
				'sku' => (string) $post_id,
			],
			'vars'        => [],
		];

		$this->assertSame( $expect, $payload );
	}

	/**
	 * Test `dmg_sailthru_content_api_data_vars` filter.
	 */
	public function test_dmg_sailthru_content_api_data_vars_filter() {
		$custom_vars = [
			'foo' => [
				'bar' => 'baz',
			],
		];

		add_filter(
			'dmg_sailthru_content_api_data_vars',
			function ( $vars ) use ( $custom_vars ) {
				return $custom_vars;
			}
		);

		$post_id = $this->factory()->post->create();

		$payload = Sync::extended_data_payload( $post_id );

		$this->assertSame( $custom_vars, $payload['vars'] );
	}

	/**
	 * Test `dmg_sailthru_content_api_data` filter.
	 */
	public function test_dmg_sailthru_content_api_data_filter() {
		$custom_data = [
			'foo' => [
				'bar' => 'baz',
			],
		];

		add_filter(
			'dmg_sailthru_content_api_data',
			function ( $data ) use ( $custom_data ) {
				return $custom_data;
			}
		);

		$post_id = $this->factory()->post->create();

		$payload = Sync::extended_data_payload( $post_id );

		$this->assertSame( $custom_data, $payload );
	}

	/**
	 * Test Sync::get_tag_slugs().
	 */
	public function test_get_tag_slugs() {
		$post_id = $this->factory()->post->create();

		// Expect an empty string for no tags
		$this->assertSame( '', Sync::get_tag_slugs( $post_id ) );

		wp_set_object_terms( $post_id, [ 'tag-1', 'tag-2', 'tag-3' ], 'post_tag', false );

		// Now expect a comma-seperated list of tag slugs
		$this->assertSame( 'tag-1,tag-2,tag-3', Sync::get_tag_slugs( $post_id ) );
	}
}
